#include <sys/poll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <errno.h>
#include <string.h>

#include "tun_dev.h"
#include "bbtun.h"
#include "bbnet.h"

BBTun::BBTun(int sock, const std::string &ip_l, const std::string & ip_r) throw (berror) : _sock(sock)
{
        _rx = 0;
	_rx_p = 0;
        _tx = 0;
	_tx_p = 0;
	_size = 0x4000;
	_buf = new char[_size];
	init(ip_l, ip_r);
}

BBTun::~BBTun() throw (berror)
{
	delete [] _buf;
	fprintf(stderr, "bbtun avg: rx/tx %.3f/%.3f\n", 
			(_rx_p)?((1. * _rx) / _rx_p):0,
			(_tx_p)?((1. * _tx) / _tx_p):0);
	close();
}

void BBTun::init(const std::string &ip_l, const std::string & ip_r) throw (berror)
{
	char dev[128]; dev[0] = 0;
	_tun = tun_open(dev);
	if (_tun < 0) {
		_tun = 0;
		throw berror_errno("Unable to open tun device: ");
	}
	if (tun_up(dev, ip_l.c_str(), ip_r.c_str(), 1200)) {
		fprintf(stderr, "Err: %s\n", strerror(errno));
		throw berror_errno("Unable to bring tun up: ");
	}
}

void BBTun::close() throw (berror)
{
	::close(_sock);
	::close(_tun);
}

void BBTun::xfer() throw (berror)
{
	struct pollfd pfd[2];
	pfd[0].fd = _sock;
	pfd[0].events = POLLIN | POLLERR;
	pfd[1].fd = _tun;
	pfd[1].events = POLLIN | POLLERR;
	int polls = ::poll(pfd, 2, 500);
	if (polls < 0)
		throw berror(std::string("Unable to poll sock/tun pair: ") + strerror(errno));
	if (!polls) {
		//fprintf(stderr, ".");
		//fflush(stderr);
		return;
	}
#if 0
	if (pfd[0].revents & POLLIN) {
		int sz = BBAds::read(_sock, _buf, _size);
                _rx += sz;
		_rx_p++;
		if (sz) BBAds::write(_tun, _buf, sz);
	}
	if (pfd[1].revents & POLLIN) {
		int sz = BBAds::read(_tun, _buf, _size);
		if (sz) BBAds::write(_sock, _buf, sz);
                _tx += sz;
		_tx_p++;
	}

#else
	if (pfd[0].revents & POLLIN) {
		int sz = BBAds::read(_sock, _buf, 2);
		if (sz != 2)
			throw berror("Unable to read 16 bits of header");
                _rx += sz;
		sz = ntohs(*((unsigned short *) _buf));
		sz &= 0x0fff; // See vtun source for description
		if (sz > _size)
			throw berror("Frame is too big");
		sz = BBAds::read(_sock, _buf + 2, sz);
                _rx += sz;
		_rx_p++;
		if (sz) BBAds::write(_tun, _buf + 2, sz);
	}
	if (pfd[1].revents & POLLIN) {
		int sz = BBAds::read(_tun, _buf + 2, _size - 2);
		*((unsigned short *) _buf) = htons(sz);
		if (sz) BBAds::write(_sock, _buf, sz + 2);
                _tx += sz;
		_tx_p++;
	}
#endif
}
